@@ -33,26 +33,23 @@
  * ========================================================================= */
 
 /* System total memory: 262144 kB RAM, 768000 kB swap
- * PID 12345: Xorg -option value
- * PID 23456: foobar -option value
- *               _______________  ______  _____________________________  ______
- * _______  __  / system memory \/system\/PID 12345                    \/PID...
- * time:  \/BL\/  used:  change:  CPU-%:  clean:  dirty: change: CPU-%:  clean:
- * 06:18:07 --   1250296        0   1.00    1234    5421       0   1.00    1234
- * 06:18:09 B-   1250300       +4   1.00    1234    5425      +4   1.00    1234
- * 06:18:11 BL   1250274      -22  10.00    1234    5421      -4 100.00    1234
- * 22:22:22 BL 999999999 -8888888 666.66 7777777 7777777 -777777 666.66 7777777
+ * PID  1547: browser
+ *                _______________  ____________  _____________________________
+ * ________  __  / system memory \/ system CPU \/PID 1547  browser         ...\
+ * time:   \/BL\/  used:  change:     %:  MHz:   clean:  dirty: change: CPU-%:
+ * 02:22:31  --   143272       +0   0.00     0    1252    2784      +0   0.00
+ * 02:22:34  --   143272       +0   1.32   253    1252    2784      +0   0.00
+ * 02:22:37  --   143332      +60   1.98   253    1252    2784      +0   0.00
+ * 02:22:40  --   143332       +0   2.30   252    1252    2784      +0   0.00
  */
 
 /* Output format without memory flags:
  *
- *            _______________  ______  _____________________________  ______
- * ________  / system memory \/system\/PID 12345                    \/PID...
- * time:   \/  used:  change:  CPU-%:  clean:  dirty: change: CPU-%:  clean:
- * 06:18:07   1250296        0   1.00    1234    5421       0   1.00    1234
- * 06:18:09   1250300       +4   1.00    1234    5425      +4   1.00    1234
- * 06:18:11   1250274      -22  10.00    1234    5421      -4 100.00    1234
- * 22:22:22 999999999 -8888888 666.66 7777777 7777777 -777777 666.66 7777777
+ *             _______________  ____________  _____________________________
+ * _________  / system memory \/ system CPU \/PID 6603  ssh root@192.168...\
+ * time:    \/  used:  change:     %:  MHz:   clean:  dirty: change: CPU-%:
+ * 12:28:32    498324       +0   0.00     2       0     460      +0   0.00
+ * 12:28:35    498084     -240   8.25   805       0     460      +0   0.00
  */
 
 #define _GNU_SOURCE
@@ -71,6 +68,7 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <fcntl.h>
+#include <math.h>
 
 #include "mem-monitor-util.h"
 
@@ -84,6 +82,30 @@ static const char watermark_high[] = "/sys/kernel/high_watermark";
 // Output to stdout by default, otherwise to file given by user.
 static FILE* output = NULL;
 
+// Bitmask holding a number of option flags
+static unsigned int 	option_flags = 0;
+
+static int 				sys_mem_change_threshold = 0;
+static float 			sys_cpu_change_threshold = 0.0f;
+
+static bool 			do_print_report_default = true;
+
+// Flags should have values of powers of 2
+enum OPTION_VALUE_FLAGS {
+	OF_PROC_MEM_CHANGES_ONLY 	= 1,
+	OF_PROC_CPU_CHANGES_ONLY 	= 2,
+	OF_SYS_MEM_CHANGES_ONLY 	= 4,
+	OF_SYS_CPU_CHANGES_ONLY 	= 8,
+	OF_INTERVAL_OPTION_SET 		= 16
+};
+// Adds a flag to the bitmask
+#define ADD_OPTION_VALUE_FLAG(option_flags_var, value) \
+	option_flags_var |= (value);
+	
+// Checks whether a flag is set in the bitmask
+#define IS_OPTION_VALUE_FLAG_SET(option_flags_var, value) \
+	( option_flags_var & value ? 1 : 0 )
+
 // Dynamic buffer to be used with getline().
 static char* dynbuf = NULL;
 static size_t dynbuf_cap = 0u;
@@ -138,6 +160,141 @@ typedef struct {
 	size_t cputicks_change;
 } monitored_process_t;
 
+
+/*
+ *  CPU stats monitoring
+ */
+
+#define CPU_STATS_SOURCE  "/sys/devices/system/cpu/cpu0/cpufreq/stats/time_in_state"
+
+/**
+ * The cpu stats data record.
+ * This structure holds how many ticks cpu has spent at the
+ * specifc frequency.
+ */
+typedef struct _cpu_stats_record_t {
+  int freq;
+  int ticks;
+  struct _cpu_stats_record_t* next;
+} cpu_stats_record_t;
+
+typedef cpu_stats_record_t* cpu_stats_t;
+
+static cpu_stats_t cpu_stats[2] = {NULL, NULL};
+static cpu_stats_t* cpu_stats_start = &cpu_stats[0];
+static cpu_stats_t* cpu_stats_end = &cpu_stats[1];
+
+/**
+ * Retrieves cpu stats record for the specified frequency.
+ *
+ * A new cpu stat record will be added, if the cpu stats holds no
+ * records for the specified frequency.
+ * @param stats[in]   the cpu stats.
+ * @param freq[in]    the cpu frequency.
+ * @return            the cpu stat record (or NULL if no records were found and
+ *                    failed to add a new one).
+ */
+static cpu_stats_record_t*
+cpu_stats_get_freq_record(cpu_stats_t* stats, int freq) {
+  cpu_stats_record_t** new_rec = stats;
+  cpu_stats_record_t* rec = *stats;
+
+  while (rec && rec->freq != freq) {
+    if (!rec->next) {
+      new_rec = &rec->next;
+      break;
+    }
+    rec = rec->next;
+  }
+  if (!(*new_rec)) {
+    *new_rec = rec = malloc(sizeof(cpu_stats_record_t));
+    if (!rec) return NULL;
+    rec->freq = freq;
+    rec->ticks = 0;
+    rec->next = NULL;
+  }
+  return rec;
+}
+
+/**
+ * Releases cpu stats data structures.
+ *
+ * @param stats[in]   the cpu stats.
+ * @return
+ */
+static void
+cpu_stats_clear(cpu_stats_t* stats) {
+  cpu_stats_record_t* rec = *stats;
+  while (rec) {
+    cpu_stats_t tmp = rec->next;
+    free(rec);
+    rec = tmp;
+  }
+  *stats = NULL;
+}
+
+/**
+ * Sets ticks value for the specified frequency.
+ *
+ * @param stats[in]   the cpu stats.
+ * @param freq[in]    the cpu frequency.
+ * @param ticks[in]   the cpu ticks spent in the specified frequency.
+ */
+static void
+cpu_stats_set_freq_ticks(cpu_stats_t* stats, int freq, int ticks) {
+  cpu_stats_record_t* record = cpu_stats_get_freq_record(stats, freq);
+  if (record) {
+    record->ticks = ticks;
+  }
+}
+
+
+/**
+ * Calculates average cpu frequency between two cpu stats snapshots.
+ *
+ * @param start[in]   the starting cpu snapshot.
+ * @param end[in]     the ending cpu snapshot.
+ * @return            the average cpu frequency.
+ */
+static int
+cpu_stats_get_avg_diff(cpu_stats_t* start, cpu_stats_t* end) {
+  cpu_stats_record_t* rec_end = *end;
+  int total_freq = 0;
+  int total_time = 0;
+  while (rec_end) {
+    cpu_stats_record_t* record = cpu_stats_get_freq_record(start, rec_end->freq);
+    if (record) {
+      int diff = rec_end->ticks - record->ticks;
+      total_time += diff;
+      total_freq += rec_end->freq * diff;
+    }
+    rec_end = rec_end->next;
+  }
+  return total_time ? total_freq / total_time : 0;
+}
+
+/**
+ * Parses the cpu stats file and creates cpu stats snapshot.
+ *
+ * @param stats[in, out]    the cpu stats.
+ * @return
+ */
+static void
+cpu_stats_take_snapshot(cpu_stats_t* stats) {
+  FILE* fp = fopen(CPU_STATS_SOURCE, "r");
+  if (fp) {
+    char line[128];
+    int freq, ticks;
+    while (fgets(line, sizeof(line), fp)) {
+      if (sscanf(line, "%d %d", &freq, &ticks) == 2) {
+        cpu_stats_set_freq_ticks(stats, freq, ticks);
+      }
+    }
+    fclose(fp);
+  }
+}
+
+
 /* Does @s1 begin with @s2?
  */
 static bool
@@ -467,7 +624,7 @@ print_headers(monitored_process_t* mprocs,
 {
 	unsigned i;
 	// First line.
-	fprintf(output, "%s           _______________  ______ ",
+	fprintf(output, "%s            _______________  ____________ ",
 			watermarks_avail ? "   " : "");
 	for (i=0; i < mprocs_cnt; ++i) {
 		fprintf(output, "%s _____________________________ %s",
@@ -475,7 +632,7 @@ print_headers(monitored_process_t* mprocs,
 	}
 	fprintf(output, "\n");
 	// Second line.
-	fprintf(output, "_______%s / system memory \\/system\\",
+	fprintf(output, "________%s / system memory \\/ system CPU \\",
 			watermarks_avail ? "  __ " : "_ ");
 	for (i=0; i < mprocs_cnt; ++i) {
 		fprintf(output, "%s/PID %-5u %-19s\\%s",
@@ -486,7 +643,7 @@ print_headers(monitored_process_t* mprocs,
 	}
 	fprintf(output, "\n");
 	// Third line.
-	fprintf(output, "time:%s\\/  used:  change:  CPU-%%:",
+	fprintf(output, "time: %s\\/  used:  change:     %%:  MHz: ",
 			watermarks_avail ? "  \\/BL" : "   ");
 	for (i=0; i < mprocs_cnt; ++i) {
 		fprintf(output, "%s  clean:  dirty: change: CPU-%%:%s",
@@ -539,7 +696,7 @@ usage()
 		"and (optionally) the status of some processes.\n"
 		"\n"
 		"Usage:\n"
-		"        %s [interval] [[PID] [PID...]]\n"
+		"        %s [OPTIONS] [interval] [[PID] [PID...]]\n"
 		"\n"
 		"Default output interval is %u seconds.\n"
 		"\n"
@@ -547,6 +704,11 @@ usage()
 		"     -f, --file=FILE       Write to FILE instead of stdout.\n"
 		"         --no-colors       Disable colors.\n"
 		"         --self            Monitor this instance of %s.\n"
+		"     -i, --interval=INTERVAL         Data acquisition interval.\n"
+		"     -C, --system-cpu-change=THRESHOLD         Perform output only when the system cpu usage is greater then the specified threshold.\n"
+		"     -M, --system-mem-change=THRESHOLD          Perform output only when the system memory change is greater then the specified threshold.\n"
+		"     -c, --cpu-changes          Perform output only when there was any change in cpu usage for any process being monitored.\n"
+		"     -m, --mem-changes          Perform output only when there was any change in memory usage for any process being monitored.\n"
 		"     -h, --help            Display this help.\n"
 		"\n"
 		"Examples:\n"
@@ -570,6 +732,11 @@ static struct option long_opts[] = {
 	{"file", 1, 0, 'f'},
 	{"pid", 1, 0, 'p'},
 	{"self", 0, 0, 1002},
+	{"mem-changes", 0, 0, 'm'},
+	{"cpu-changes", 0, 0, 'c'},
+	{"system-cpu-change", 1, 0, 'C'},
+	{"system-mem-change", 1, 0, 'M'},
+	{"interval", 1, 0, 'i'},
 	{0,0,0,0}
 };
 
@@ -582,7 +749,7 @@ parse_cmdline(int argc, char** argv,
 	char* output_path = NULL;
 	unsigned cnt=0, cap=0;
 	monitored_process_t* m = NULL;
-	while ((opt = getopt_long(argc, argv, "p:hf:", long_opts, NULL)) != -1) {
+	while ((opt = getopt_long(argc, argv, "p:hf:mcM:C:i:", long_opts, NULL)) != -1) {
 		switch (opt) {
 		case 'f':
 			output_path = strdup(optarg);
@@ -599,6 +766,33 @@ parse_cmdline(int argc, char** argv,
 		case 'p':
 			monitor_pid(atoi(optarg), &m, &cnt, &cap);
 			break;
+		case 'm':
+			ADD_OPTION_VALUE_FLAG(option_flags, OF_PROC_MEM_CHANGES_ONLY);
+			break;
+		case 'c':
+			ADD_OPTION_VALUE_FLAG(option_flags, OF_PROC_CPU_CHANGES_ONLY);
+			break;
+		case 'M':
+			ADD_OPTION_VALUE_FLAG(option_flags, OF_SYS_MEM_CHANGES_ONLY);
+			sys_mem_change_threshold = atoi(optarg);
+			do_print_report_default = false;
+			break;
+		case 'C':
+			ADD_OPTION_VALUE_FLAG(option_flags, OF_SYS_CPU_CHANGES_ONLY);
+			if (!sscanf(optarg, "%f", &sys_cpu_change_threshold) ) {
+				fprintf(stderr, "ERROR: invalid CPU change threshold for the system\n");
+				exit(1);
+			}
+			do_print_report_default = false;
+			break;
+		case 'i':
+			if (sscanf(optarg, "%u", sleep_interval) != 1 ||
+				sleep_interval == 0u) {
+				fprintf(stderr, "ERROR: invalid interval\n");
+				exit(1);
+			}
+			ADD_OPTION_VALUE_FLAG(option_flags, OF_INTERVAL_OPTION_SET);
+			break;
 		default:
 			usage();
 			exit(1);
@@ -616,11 +810,19 @@ parse_cmdline(int argc, char** argv,
 		output = stdout;
 	}
 	if (optind < argc) {
-		if (sscanf(argv[optind], "%u", sleep_interval) != 1 ||
-		    sleep_interval == 0u) {
+		// read interval argument
+		unsigned _interval = 0;
+		if (sscanf(argv[optind], "%u", &_interval) != 1 ||
+			sleep_interval == 0u) {
 			fprintf(stderr, "ERROR: invalid interval\n");
 			exit(1);
 		}
+		// only overwrite interval if interval is not already specified by -i option
+		if (!IS_OPTION_VALUE_FLAG_SET(option_flags, OF_INTERVAL_OPTION_SET) ) {
+			 *sleep_interval = _interval;
+		} else {
+			fprintf(stderr, "WARNING: interval argument '%u' is ignored as '-i option' is specified\n", _interval);
+		}
 		++optind;
 	}
 	while (optind < argc) {
@@ -631,6 +833,14 @@ parse_cmdline(int argc, char** argv,
 		*mprocs = m;
 		*mprocs_cnt = cnt;
 	}
+	
+	// determine if no printing needs to be done by default 
+	if (*mprocs_cnt > 0 && 
+		(IS_OPTION_VALUE_FLAG_SET(option_flags, OF_PROC_MEM_CHANGES_ONLY) || 
+		 IS_OPTION_VALUE_FLAG_SET(option_flags, OF_PROC_CPU_CHANGES_ONLY) ) ) {
+
+		do_print_report_default = false;
+	}
 }
 
 /* When printing results to console, we want to periodically reprint the
@@ -684,19 +894,19 @@ mem_flags(bool watermarks_avail)
 	const bool flag_low  = check_flag(watermark_low);
 	const bool flag_high = check_flag(watermark_high);
 	if (flag_low && flag_high) {
-		if (colors) return COLOR_HIGHMARK "BL" COLOR_CLEAR;
-		else        return "BL";
+		if (colors) return COLOR_HIGHMARK " BL" COLOR_CLEAR;
+		else        return " BL";
 	}
 	if (flag_low) {
-		if (colors) return COLOR_LOWMARK "B-" COLOR_CLEAR;
-		else        return "B-";
+		if (colors) return COLOR_LOWMARK " B-" COLOR_CLEAR;
+		else        return " B-";
 	}
 	if (flag_high) {
 		// Only highmark set? Should not happen...
-		if (colors) return COLOR_HIGHMARK "-L" COLOR_CLEAR;
-		else        return "-L";
+		if (colors) return COLOR_HIGHMARK " -L" COLOR_CLEAR;
+		else        return " -L";
 	}
-	return "--";
+	return " --";
 }
 
 int main(int argc, char** argv)
@@ -710,6 +920,8 @@ int main(int argc, char** argv)
 	unsigned mprocs_cnt = 0;
 	bool watermarks_avail = false, is_atty = false;
 	unsigned rows=0, lines_printed=0;
+	cpu_stats_t* cpu_stats_swap;
+
 	parse_cmdline(argc, argv, &mprocs, &mprocs_cnt, &sleep_interval);
 	(void) nice(-19);
 	if (!system_memory_totals(&ram_total, &swap_total)) {
@@ -737,6 +949,7 @@ int main(int argc, char** argv)
 	if (!is_atty) colors = false;
 	fprintf(output, "System total memory: %u kB RAM, %u kB swap\n",
 			ram_total, swap_total);
+	
 	(void) print_process_names(mprocs, mprocs_cnt);
 	lines_printed += print_headers(mprocs, mprocs_cnt, watermarks_avail);
 	// Disable header reprinting if we're printing to console, or if the
@@ -745,33 +958,87 @@ int main(int argc, char** argv)
 	// Install our signal handler, unless someone specifically wanted
 	// SIGINT to be ignored.
 	if (signal(SIGINT, quit_app) == SIG_IGN) signal(SIGINT, SIG_IGN);
+	
+	bool do_print_report;
+	
+	unsigned sys_ram_used_last_printed = ram_used;
+	float sys_cpu_usage_last_printed = cpu_usage(cpu_ticks_total-cpu_ticks_total_prev, cpu_ticks_idle-cpu_ticks_idle_prev);
+	
+  /* initialize cpu stats */
+  cpu_stats_take_snapshot(cpu_stats_start);
+
 	while (!quit) {
+
+		do_print_report = do_print_report_default;
+		
 		const time_t t = time(NULL);
 		const struct tm* ts = localtime(&t);
 		if (!ts) {
 			fprintf(stderr, "ERROR: localtime() failed\n");
 			return 1;
 		}
-		fprintf(output,
-			"%02u:%02u:%02u %s%9u %+8d %6.2f",
-			ts->tm_hour, ts->tm_min, ts->tm_sec,
-			mem_flags(watermarks_avail),
-			ram_used, (int)ram_used - (int)prev_ram_used,
-			cpu_usage(cpu_ticks_total-cpu_ticks_total_prev,
-			          cpu_ticks_idle-cpu_ticks_idle_prev));
-		for (unsigned i=0; i < mprocs_cnt; ++i) {
-			fprintf(output, "%s %7u %7u %+7d %6.2f%s",
-				c_begin(i),
-				mprocs[i].mem_clean,
-				mprocs[i].mem_dirty,
-				mprocs[i].mem_change,
-				cpu_usage(cpu_ticks_total-cpu_ticks_total_prev,
-				          (cpu_ticks_total-cpu_ticks_total_prev) -
-				           mprocs[i].cputicks_change),
-				c_end(i));
+		// check system for changes in mem and cpu
+		int _sys_ram_change = (int)ram_used - (int)sys_ram_used_last_printed;
+		float _sys_cpu_usage = cpu_usage(cpu_ticks_total-cpu_ticks_total_prev, cpu_ticks_idle-cpu_ticks_idle_prev);
+		float _sys_cpu_usage_change = _sys_cpu_usage - sys_cpu_usage_last_printed; 
+		
+		if ( (IS_OPTION_VALUE_FLAG_SET(option_flags, OF_SYS_MEM_CHANGES_ONLY) && abs(_sys_ram_change) >= sys_mem_change_threshold) || 
+			(IS_OPTION_VALUE_FLAG_SET(option_flags, OF_SYS_CPU_CHANGES_ONLY) && fabs(_sys_cpu_usage_change) >= sys_cpu_change_threshold) ) {
+
+			do_print_report = true;
 		}
-		fprintf(output, "\n");
-		fflush(output);
+		
+		if ( !do_print_report ) {
+			// check all monitored processes for changes in mem and cpu
+			for (unsigned i = 0; i < mprocs_cnt; ++i) {
+				if ( (IS_OPTION_VALUE_FLAG_SET(option_flags, OF_PROC_MEM_CHANGES_ONLY) && mprocs[i].mem_change != 0) ||
+					 
+					 (IS_OPTION_VALUE_FLAG_SET(option_flags, OF_PROC_CPU_CHANGES_ONLY) && mprocs[i].cputicks_change != 0.0f) ) {
+					
+					do_print_report = true;
+					break;
+				}
+			}
+		}
+			
+		// print report for the system
+		if (do_print_report) {
+			fprintf(output,
+				"%02u:%02u:%02u %s%9u %+8d %6.2f",
+				ts->tm_hour, ts->tm_min, ts->tm_sec,
+				mem_flags(watermarks_avail),
+				ram_used, _sys_ram_change, _sys_cpu_usage);
+
+	    /* Print cpu stats. */
+	    cpu_stats_take_snapshot(cpu_stats_end);
+	    fprintf(output, "%6d", cpu_stats_get_avg_diff(cpu_stats_start, cpu_stats_end) / 1000);
+	    /* Because cpu stats can only grow we can simply swap between the start and end stats.
+	     * The end stats will become the start stats and the old start stats data structures will be used
+	     * to take new cpu stats snapshot, making it the new end stats.
+	     */
+	    cpu_stats_swap = cpu_stats_end;
+	    cpu_stats_end = cpu_stats_start;
+	    cpu_stats_start = cpu_stats_swap;
+				
+			sys_ram_used_last_printed = ram_used;
+			sys_cpu_usage_last_printed = _sys_cpu_usage;
+			
+			// print report for the processes
+			for (unsigned i=0; i < mprocs_cnt; ++i) {
+				fprintf(output, "%s %7u %7u %+7d %6.2f%s",
+					c_begin(i),
+					mprocs[i].mem_clean,
+					mprocs[i].mem_dirty,
+					mprocs[i].mem_change,
+					cpu_usage(cpu_ticks_total-cpu_ticks_total_prev,
+							  (cpu_ticks_total-cpu_ticks_total_prev) -
+							   mprocs[i].cputicks_change),
+					c_end(i));
+			}
+			fprintf(output, "\n");
+			fflush(output);
+		}
+				
 		sleep(sleep_interval);
 		if (quit) break;
 		prev_ram_used = ram_used;
@@ -779,14 +1046,18 @@ int main(int argc, char** argv)
 			fprintf(stderr, "ERROR: unable to read /proc/meminfo\n");
 			return 1;
 		}
+		
 		cpu_ticks_total_prev = cpu_ticks_total;
 		cpu_ticks_idle_prev  = cpu_ticks_idle;
 		system_cpu_usage(&cpu_ticks_total, &cpu_ticks_idle);
 		update_processes(mprocs, mprocs_cnt);
-		if (is_atty && rows) {
-			if (++lines_printed >= rows-1) {
-				lines_printed = print_headers(mprocs,
-						mprocs_cnt, watermarks_avail);
+		
+		if (do_print_report) {
+			if (is_atty && rows) {
+				if (++lines_printed >= rows-1) {
+					lines_printed = print_headers(mprocs,
+							mprocs_cnt, watermarks_avail);
+				}
 			}
 		}
 	}
@@ -797,6 +1068,10 @@ int main(int argc, char** argv)
 	}
 	free(mprocs);
 	free(dynbuf);
+
+	/* release cpu stats data */
+	cpu_stats_clear(&cpu_stats[0]);
+	cpu_stats_clear(&cpu_stats[1]);
 }
 
 /* ========================================================================= *
